/*$$
 * Template for a 2D Repast model -- Agent description
 *   M.Schippling 6/6/6
 * Based on the SugarScape demo from
 * Copyright (c) 1999, Trustees of the University of Chicago
 *$$*/
package com.etantdonnes.repast.models.template2D;

import uchicago.src.sim.gui.Drawable;
import uchicago.src.sim.gui.SimGraphics;
import uchicago.src.sim.util.Random;
import cern.jet.random.Uniform;

/**
 * The agent for the simulation.
 * The source is annotated so see that for more info.
 *
 * @author Nick Collier, MSchippling
 * original version:
 * @version $Revision: 1.1 $ $Date: 2005/08/12 20:04:53 $
 */

// If an agent, or any other object is to be displayed in a display it
// needs to implement the Drawable interface method: draw(SimGraphics g)

public class T2DAgent implements Drawable
{
	protected int		x, y;	// current coordinates in space
	protected T2DSpace	space;	// back links to our space and model objects
	protected T2DModel	model;

	/** create a new agent in given model
	 * @param ss	-- working space
	 * @param mm	-- model
	 */
	public T2DAgent( T2DSpace ss, T2DModel mm )
	{
		this.space = ss;
		this.model = mm;
	}

	/** convenient accessors for position in space
	 */
	public void setXY( int x, int y )
	{
		this.x = x;
		this.y = y;
	}
	
	/** required Drawable implementation
	 */
	public int getX()
	{
		return x;
	}

	/** required Drawable implementation
	 */
	public int getY()
	{
		return y;
	}

	/** required Drawable implementation
	 * draw() is called whenever the display is updated,
	 *  assuming the agent is part of what is being displayed.
	 */
	public void draw( SimGraphics g )
	{
		// model specific, mark each agent as a red blot
		g.drawFastRoundRect( java.awt.Color.red );
	}
	
	/** step() is called by the scheduler once per tick,
	 *   to do whatever is necessary for each agent.
	 *   See Model.java for how the scheduler gets setup and this gets called.
	 *     step() is an arbitrary but convenient name...
	 */
	public void step()
	{
		// this all model specific stuff
		
		// this dumb model just has agents that do a random walk
		//   and don't collide with each other or -1's in the space
		
		// figure out which direction to move
		int dir;
		int rval;
		int newx = x;	// use local copies because we update position
		int newy = y;	//  using moveAgent()...
		
		do
		{
			// move in the great circle pattern:
			//		3 2 1
			//		4 A 0
			//		5 6 7
			
			dir = Random.uniform.nextIntFromTo(0, 7);
			switch( dir )
			{
				case 0:		++newx;     	break;	
				case 1:		++newx; ++newy;	break;
				case 2:		        ++newy;	break;
				case 3:		--newx; ++newy;	break;
				case 4:		--newx;     	break;
				case 5:		--newx; --newy;	break;
				case 6:		        --newy;	break;
				case 7:		++newx; --newy;	break;
				default:
			}
			
			// move ourself in the space, but only to unoccupied locations
			//  this also takes care of torus wrap-around in agent.x,y
			// keep track of collisions just to have something to graph
			rval = space.moveAgent( this, newx, newy );
			if( rval == T2DSpace.ESPACE )
				++collideSpace;
			else if( rval == T2DSpace.EAGENT )
				++collideAgent;
			
		} while( rval != T2DSpace.AOK );
		
		// save which way we went for logging and stuff...
		direction = dir;
	}
	
	// model specific stuff
	
	// Just as the model needs get and set accessor methods in order for
	// its initial parameters to be displayed and thus subject to modification,
	// an agent (or any other object) can be probed through similar get and set
	// methods.
	//
	// Probing consists of clicking on an object in the display, causing that
	// object's current state to be displayed. What is displayed depends on
	// the various get and set methods implemented by the object. For example,
	// if an object has setDirection() and getDirection() methods, a Direction
	// field will be displayed providing its current value and allowing the
	// user to enter a new value and press enter.
	// As of the 3.1 release only number, String, and boolean fields can be
	//  displayed. Of course a user can use the get and set methods
	// to turn a Vector, for example, into a String or whatever is appropriate.
	
	// for this dumb model we keep track of the followng just to make graphs
	int direction;			// which way we went last
	int collideSpace = 0;	// count of collisions with space boundaries
	int collideAgent = 0;	// count of collisions with other agents

	/**
	 * @param d  set the last direction value.
	 */
	public void setDirection( int d )
	{
		direction = d;
	}

	/**
	 * @return the last direction value.
	 */
	public int getDirection()
	{
		return direction;
	}
	
	/**
	 * @return the collideAgent count.
	 */
	public int getCollideAgent()
	{
		return this.collideAgent;
	}

	/**
	 * @param collideAgent The collideAgent count to set.
	 */
	public void setCollideAgent( int collideAgent )
	{
		this.collideAgent = collideAgent;
	}

	/**
	 * @return the collideSpace count.
	 */
	public int getCollideSpace()
	{
		return this.collideSpace;
	}

	/**
	 * @param collideSpace The collideSpace count to set.
	 */
	public void setCollideSpace( int collideSpace )
	{
		this.collideSpace = collideSpace;
	}

} // end'o'T2DAgent
