/*
 * $$ Template for a 2D Repast model -- Model description
 *  M.Schippling 6/6/6
 * Based on the SugarScape demo from Copyright (c) 1999, Trustees of the
 * University of Chicago $$
 */
package com.etantdonnes.repast.models.template2D;

import java.util.ArrayList;
import java.util.ListIterator;
import java.util.Vector;

import uchicago.src.reflector.BooleanPropertyDescriptor;
import uchicago.src.sim.analysis.DataRecorder;
import uchicago.src.sim.analysis.DataSource;
import uchicago.src.sim.analysis.OpenHistogram;
import uchicago.src.sim.analysis.OpenSequenceGraph;
import uchicago.src.sim.analysis.OpenStats;
import uchicago.src.sim.analysis.Sequence;
import uchicago.src.sim.engine.Schedule;
import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.SimModelImpl;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.gui.Object2DDisplay;
import uchicago.src.sim.gui.Value2DDisplay;
import uchicago.src.sim.util.SimUtilities;
import cern.jet.random.Uniform;

/**
 * The source code is heavily annotated as example of a simulation built using
 * the Repast toolkit. See the html API documentation for the details of the
 * framework objects.
 * @author Nick Collier, M.Schippling
 *   original version:
 * @version $Revision: 1.1 $ $Date: 2005/08/12 20:04:53 $
 */

// Simulation models should extend SimModelImpl which does some basic setup
// and allows a controller easy access to a simulations initial parameters.
public class T2DModel extends SimModelImpl
{
	// Every model must have a schedule
	protected Schedule		schedule;

	// The working Space - the agent's environment
	//   grid values are initialized from a file
	protected T2DSpace		space;
	
	// our Agents
	protected int			numAgents;	// starting number of Agents to create
	protected ArrayList		agentList;	// list of agents to iterate over

	// The surface on which the agents and space are displayed
	protected DisplaySurface		dsurf;

	// A graph that plots sequences
	protected OpenSequenceGraph	graph;
	// A histogram
	protected OpenHistogram		hist;

	// A DataRecorder allows data generated by a simulation
	//   to be written to a file. See use examples below...
	// writelog enables logging
	private boolean			writelog = false;
	protected DataRecorder	recorder;

	/** utility main() for standalone running
	 */
	public static void main( String[] args )
	{
		SimInit init = new SimInit();
		T2DModel model = new T2DModel();
		init.loadModel( model, "", false );
	}

	// All simulations should have a no argument constructor -
	//   this is empty and thus not strictly necessary -- err...wot?
	public T2DModel()
	{
		// I have no idea what this does or why...schip
		BooleanPropertyDescriptor bd = new BooleanPropertyDescriptor(
				"Replacement", false );
		descriptors.put( "Replacement", bd );

	}

	/** required method -
	 *    name to display on the Controller toolbar.
	 */
	public String getName()
	{
		return "Template";
	}

	// Every model must have begin() and setup() methods
	//  (required for implementing the SimModel inteface).

	/** begin() intializes the model for the start of a run. Consequently,
	** the build* methods are called here, and any displays are displayed.
	** This is called whenever the start button is clicked,
	*   or the step button if the run has not yet started.
	*   
	*  note that this is called after setup and can load values from
	*  GUI screens into various local variables...
	*/
	public void begin()
	{
		// call our convenience methods to set it up
		buildModel();
		buildDisplay();
		buildSchedule();

		// display what we want to display
		dsurf.display();
		graph.display();
		hist.display();
	}

	/** setup() prepares the model for another run.
	**  This is called when the setup button is clicked.
	**  Clean up previous run -- set any objects that are created over the
	**    course of a run to null, and dispose of any DisplaySurfaces or graphs.
	**    While not strictly necessary, calling System.gc() might prevent
	**     memory leaks.
	** Then create new schedule and display objects and
	**   initialize parameters to whatever defaults the user wants.
	**/
	public void setup()
	{
		// ##### get rid of old stuff #####
		
		schedule = null;
		agentList = null;
		space = null;
		
		recorder = null;	// I don't think there is any close() to call...
		
		if( dsurf != null )
			dsurf.dispose();
		dsurf = null;
		
		// example graphs
		if( graph != null )
			graph.dispose();
		graph = null;

		if( hist != null )
			hist.dispose();
		hist = null;

		// do model specific cleanup

		// try to get a clean slate
		System.gc();
		
		// ##### create new stuff #####

		// create a schedule with an interval of one.
		schedule = new Schedule( 1 );
		
		// get an empty agent list
		agentList = new ArrayList();
		
		// make a new display window
		dsurf = new DisplaySurface( this, getName() );
		
		// By registering a DisplaySurface, you are allowing Repast to automate
		// some tasks on the displays surface, such as making movies or taking
		// snapshots.
		registerDisplaySurface( getName(), dsurf );

		// examples of graphs....
		
		// creates a dynamic histogram to display in a window:
		//  called Histogram, with 8 bins, and a lowerbound of -1
		//   (note: LB==-1 is because of fan-hitting in the histogram lib
		//    when everything is at the LB {0LB and 0maxval}...bug or no?)
		//  the data source for this is set in buildDisplay()
		hist = new OpenHistogram( "Histogram", 8, -1, this );
		this.registerMediaProducer( "Hist", hist );

		// Create a sequence graph:
		//  called Graph,
		//  allow the plotted data to be written to the file ./graph_data.txt
		//    in comma delimited format.
		//  the data source(s) for this is set in buildDisplay()
		graph = new OpenSequenceGraph( "Graph", this,
				"./graph_data.txt", OpenStats.CSV );
		this.registerMediaProducer( "Graph", graph );
		
		// initialize default values which may be modified by the GUI
		//   before the begin() and the build's() are called

		// agent properties
		numAgents = 20;
		
		// initialize model specific stuff here...
	}


	// The typical way to code a simulation is to divide up the creation of the
	// simulation into three methods - buildModel(), buildDisplay(),
	// buildSchedule(). This division is not strictly necessary, but it does
	// make the creation conceptually clearer.

	/** The buildModel() method is responsible for creating those parts of the
	** simulation that represent what is being modeled. Consequently, the agents
	** and their environment are typically created here together with any
	** optional data collection objects. Of course, this method may call other
	** methods to help build the model.
	**/
	private void buildModel()
	{
		// create the space using the values in template.pgm as
		// the initialized max vaues (see T2DSpace.java for more).
		space = new T2DSpace(
				"/com/etantdonnes/repast/models/template2D/template.pgm" );

		// make the agents,
		//  note that addNewAgent() places them in the space.agentGrid
		//    and is basically model-specific
		for( int i = 0; i < numAgents; i++ )
		{
			this.addNewAgent();
		}
		
		// if we asked to log data, create all the necessary objects
		if( writelog )
		{
			// create a new DataRecorder to log
			//  the sim data to ./[name]_data.txt
			recorder = new DataRecorder( "./" + getName() + "_data.txt", this );
			
			// Add ObjectDataSources to the recorder as a source of data.
			// Here we use external inner classes for logging so we can share
			//  the class defs with the graphs that we have instantiated.
			// See the specified classes for details...
			recorder.addObjectDataSource(	"SpaceCollision",
											new spaceCollisions() );
			
			recorder.addObjectDataSource(	"AgentCollision",
											new agentCollisions() );
			
			// Adds a data source that logs the count of the agents alive
			// in the simulation.
			// This uses a local inner class that implements DataSource,
			//  instead of external classes as above...
			class AgentCountClass implements DataSource
			{
				// whatever gets returned gets toString()ed into the file...
				public Object execute()
				{
					int i = agentList.size();
					return new Integer( i );
				}
			};
			
			// Add above class to the logging list
			recorder.addObjectDataSource(	"AgentCount",
											new AgentCountClass() );
		}
		
	}

	/** buildDisplay() builds those parts of the simulation that
	**  have to do with displaying the simulation to a user.
	**/
	private void buildDisplay()
	{
		// Create a display to show the agentGrid on the screen.
		// This will display any Drawable contained within the Object2DTorus.
		// In a non-batch simulation the agents in a simulation will typically
		// implement the Drawable interface.
		// note that both the agent and space Displays are sub-classes
		//   of Display2D which is the basic Object that we need here...
		
		// get the stuff to display for the agents
		Object2DDisplay agentDisplay = space.getAgentDisplay();

		// Rather than have the agentDisplay iterating over each object in the
		// Object2DTorus, give the display a list of Drawables to show.
		//  This agentList is also used in all the graph and log code.
		agentDisplay.setObjectList( agentList );
		
		// Get the stuff to display for the space
		Value2DDisplay spaceDisplay = space.getSpaceDisplay();

		// DisplaySurface does the actual drawing of the various displays on the
		// user's screen. Displays are displayed in the order they are added.
		// Our dsurf object is created in the setup() method
		dsurf.addDisplayableProbeable( spaceDisplay, getName() + "Space" );
		dsurf.addDisplayableProbeable( agentDisplay, getName() + "Agents" );

		// This causes the display surface to update the display whenever the
		// simulation run is paused or ended. The DisplaySurface now listens
		// for SimEvents produced by the T2DModel
		addSimEventListener( dsurf );

		
		// setup graphs
		
		// Adds two Sequences to be displayed on the graph.
		//  These Squences will show the values returned by the given classes,
		//  labled with the given names and in arbitrarily selected colors.
		graph.addSequence( "SpaceCollisions", new spaceCollisions() );
		graph.addSequence( "AgentCollisions", new agentCollisions() );

		// set some convenient graph parameters
		graph.setAxisTitles( "Time", "Number" );
		graph.setXRange( 0, 50 );
		graph.setYRange( 0, 10 );
		graph.setSize( 400, 250 );

		// Histograms are generated from HistogramItems. The HGI iterates
		//   over the given List each time hist.step() is called and generates
		//   the display.
		//  Below, creates an item called Direction that gets its data
		//   by calling getDirection() on each agent in the agentList.
		//   The last two args are the digits before and after the decimal pt.
		// Note: according to comments in the Hist..class one should really use
		//   the interface that uses a BinDataSource instead of a reflected
		//   method call. This would make abstracting the histogram generation
		//   a little clearer perhaps, but I'm too lazy to change right now...
		// Rant: this creates a histogram for all the given values in
		//   one single step of the model. There doesn't seem to be a way to
		//   accumulate over multiple steps to generate a histogram for the
		//   entire model run, and, in fact, it appears that this would
		//   require hacking the very LOWEST level of the histogram code.
		//   This is a clearcut example of something I want to do being
		//    designed out of the system. I don't know how this happens...
		hist.createHistogramItem( "Direction", agentList,
								  "getDirection", 1, 0 );
		hist.setYRange( 0, 100.0 );
	}

	/** buildSchedule() builds the schedule that changes the simulation state.
	** Under this scheme, a simulation is a state machine where all transitions
	** between states are the result of actions initiated by a schedule.
	**/
	private void buildSchedule()
	{
		// make our step() method get called on every tick
		schedule.scheduleActionBeginning(0, this, "step");
		
		// On a pause or end in the simulation run, call the writeToFile
		//  method on the recorder object to dump stuff to the log.
		if( writelog )
		{
			schedule.scheduleActionAtPause( recorder, "writeToFile" );
			schedule.scheduleActionAtEnd( recorder, "writeToFile" );
		}
	}

	/** What to do at every timeclick.
	 *  Not a required method, but convenient to locate all the action...
	 *  Gets scheduled by buildSchedule() and must be public for that to work.
	 */
	public void step()
	{
		// do model specific pre-step work
		
		// just to keep everyone on their toes,
		//  this will step() the agents in a random order each time.
		shuffleAgents();

		// call the step method on each Agent in the agentList
		for( int i = 0; i < agentList.size(); i++ )
		{
			T2DAgent agent = (T2DAgent) agentList.get( i );
			agent.step();
		}
		
		// do whatever is necessary to the space
		space.step();

		// Update display after things change their state.
		//  Otherwise, display won't be in synch at end or pause.
		dsurf.updateDisplay();
		graph.step();
		hist.step();
		
		// log the necessary...
		if( writelog )
			recorder.record();
		
		// do model specific post-step work
		
	}
	
	/** convenient accessors and manipulators **/
	
	/** required method -
	 *   names of parameters that can be modified,
	 *   will be listed in alpha order in Settings window.
	 *   Each param needs get/setter too...
	 *   See the String params[] def below for actual data.
	 */
	public String[] getInitParam()
	{
		return params;
	}

	/** required accessor method
	 *   @return -- the current schedule
	 */
	public Schedule getSchedule()
	{
		return schedule;
	}

	/** Randomize the order of the Agents in the agentList
	 *   Just for convenience to make interactions less regular
	 */
	protected void shuffleAgents()
	{
		SimUtilities.shuffle( agentList );
	}

	/**  Create an agent and place it in a random but unoccupied location.
	 *    This is almost all model specific, but everyone needs agents...
	 */
	public void addNewAgent()
	{
		// make a new one
		T2DAgent agent = new T2DAgent( space, this );
		
		// add agent to our current list
		agentList.add( agent );

		// model specific agent initialization stuff
		
		// find an empty grid location and plop it there...
		space.placeAgent( agent );
		
		// do whatever else the model says to do with a new agent
	}
	
	// When a simulation is started through SimInit, some BaseController is
	// created to control the running of that model. This BaseController calls
	// getInitParam() on the model and receives a list of initial parameters
	// that can be displayed for modification to the user. In order to display
	// the value of these parameters the controller determines if the model
	// has implemented get and set methods for that parameter. If so, the
	// controller calls the get method and displays the result to the user. A
	// similar process occurs when a model's initial starting parameters
	// are written to a data file.
	//
	// What this means is that if a user wants to display some initial starting
	// parameter and have this parameter be modifiable, the name of the
	// parameter must be returned by the getInitParam() method and the model
	// must contain the appropriate get and set methods. For example, to do
	// the above for a parameter called numAgents, "NumAgents" must be present
	// in the array returned by getInitParam and this model class must have a
	// getNumAgents() method and a setNumAgents() method. The parameter name
	// must match the method names minus the "get" and "set",
	//  e.g.: "numAgents" and getNumAgents().
	//
	// The following methods are an example of this pattern.
	
	/** initial number of agents
	 * @param num -- number of agents to create at startup.
	 */
	public void setNumAgents( int num )
	{
		numAgents = num;
	}
	
	public int getNumAgents()
	{
		return numAgents;
	}
	
	/**
	 * @param l -- turns on file logging.
	 */
	public void setWritelog( boolean l )
	{
		this.writelog = l;
	}

	/**
	 * @return -- logging state.
	 */
	public boolean getWritelog()
	{
		return this.writelog;
	}

	// model specific stuff
	
	// names of parameters to be displayed in Settings window
	String[] params =
	{ "NumAgents", "Writelog", // almost always....
		// other model specific init params:
	};
	
	// These inner classes are used by all the data collection methods - i.e.
	// DataRecorder, SequenceGraph and the Histogram. They are here as examples.

	// Creates a DataSource and Sequence object that returns the total
	//  number of space collisions for all the agents in the agent list.
	class spaceCollisions implements DataSource, Sequence
	{
		/** the DataSource impl
		 * @return -- value to log, as Object
		 */
		public Object execute()
		{
			return new Double( getSValue() );
		}

		/** the Sequence impl
		 * @return -- value to plot on graph, etal
		 */
		public double getSValue()
		{
			int val = 0;
			for( int i = 0; i < agentList.size(); i++ )
			{
				T2DAgent agent = (T2DAgent) agentList.get( i );
				val += agent.getCollideSpace();
			}

			return val;
		}
	};

	// like above but for agent collisions
	class agentCollisions implements DataSource, Sequence
	{
		/** the DataSource impl
		 * @return -- value to log, as Object
		 */
		public Object execute()
		{
			return new Double( getSValue() );
		}

		/** the Sequence impl
		 * @return -- value to plot on graph, etal
		 */
		public double getSValue()
		{
			int val = 0;
			for( int i = 0; i < agentList.size(); i++ )
			{
				T2DAgent agent = (T2DAgent) agentList.get( i );
				val += agent.getCollideAgent();
			}

			return val; // only one agent knows it tried to collide...
		}
	};

} // end'o'T2DModel