/*$$
 * Template for a 2D Repast model -- Space description
 *   M.Schippling 6/6/6
 * Based on the SugarScape demo from
 * Copyright (c) 1999, Trustees of the University of Chicago
 *$$*/
package com.etantdonnes.repast.models.template2D;

import java.awt.Color;
import java.awt.Dimension;

import uchicago.src.collection.BaseMatrix;
import uchicago.src.sim.gui.ColorMap;
import uchicago.src.sim.gui.Object2DDisplay;
import uchicago.src.sim.gui.Value2DDisplay;
import uchicago.src.sim.space.Object2DGrid;
import uchicago.src.sim.space.Object2DTorus;
import cern.jet.random.Uniform;

/**
 * The space in which the Agents act.
 * Gets initialized to a size and pattern from the template.pgm file.
 * 
 * The source has been annotated so see that for further details.
 *
 * @author Nick Collier, M.Schippling
 *   original version:
 * @version $Revision: 1.1 $ $Date: 2005/08/12 20:04:53 $
 */

public class T2DSpace
{
	// the model space is composed of three tori,
	//  1) is initialized from a data file (as an example in the template)
	//  2) is the current environment value at each x,y coordinate.
	//  3) contains the agents just so we can find them and display them.
	// locations are matched between the space and agent tori as the model runs
	protected Object2DTorus	initSpace;		// space initialized from data file
	protected Object2DTorus	currentSpace;	// working space
	protected Object2DTorus	agentGrid;		// where the agents live
	private Dimension		size;	// the space dimensions, for convenience

	public T2DSpace( String initFile )
	{
		//  Note that all these spaces _could_ be in a Multi2DTorus
		//   but that would make it too obvious...and there isn't an init()
		//   constructor to get stuff from the file.
		
		// Construct a torus initialized using the values in the given file.
		// we get this as a stream so can load it from a jar file.
		// if we didn't need to get this from a jar we could just pass
		// the file name to Object2DTorus.
		java.io.InputStream stream = getClass().getResourceAsStream( initFile );
		initSpace = new Object2DTorus( stream, Object2DGrid.PGM_ASCII );

		// save dimensions for convenience
		size = initSpace.getSize();
		//System.out.println( size.width + ", " + size.height );
		
		// make an empty working space with the same size as initSpace
		//  this will be used to hold information about the environment
		currentSpace = new Object2DTorus( size.width, size.height );
		
		// make an agent living space,
		//  each location in the grid will map onto a currentSpace location
		agentGrid = new Object2DTorus( size.width, size.height );

		// initialize details of model spaces,
		// sets the currentSpace locations to the initSpace values.
		// Note: this is really a no-op step because we never refer back
		//   to the initSpace, but it shows how to copy spaces anyway...
		//  Rant: Strangely, there is no copy constructor for Spaces, so
		//   we gotta do a full iteration through the thing... I guess
		//   making sure clone() works on the contained objects is tough.
		BaseMatrix m = initSpace.getMatrix();
		for( int i = 0; i < m.getNumCols(); i++ )
		{
			for( int j = 0; j < m.getNumRows(); j++ )
			{
				Integer x = (Integer) m.get( i, j );
				currentSpace.putObjectAt( i, j, new Integer( x.intValue() ) );
			}
		}

	}

	/** convenient accessors
	 */
	public Dimension getSize()
	{
		return size;
	}

	public int getXSize()
	{
		return size.width;
	}

	public int getYSize()
	{
		return size.height;
	}

	public Object2DGrid getCurrentSpace()
	{
		return currentSpace;
	}
	
	
	/** Model particulars
	 **/
	
	/** Where the action is...do the necessary on each model step
	 */
	public void step()
	{
		// all model specific stuff
		//  and this model has no specifics...
	}
	
	/* some convenient agent position manipulators
	 * pretty much generally useful, but still model specific
	 */
	
	/**
	 * return the agent at the specified location
	 */
	public T2DAgent getAgentAt( int x, int y )
	{
		return (T2DAgent) agentGrid.getObjectAt( x, y );
	}

	/** Put the specified agent in the specified location if it is free.
	 * adjusts actual agent location for torus wrap-around
	 * check for already occupied and return appropriately.
	 * To clear a spot, put a null agent.
	 * 
	 * @return -- indicator of success or type of failure
	 *				AOK = 0;		-- put successful
	 *				ESPACE = 1;		-- failed due to overlap with space
	 *				EAGENT = 2;		-- failed due to overlap with another agent
	 */
	public static final int AOK = 0;
	public static final int ESPACE = 1;
	public static final int EAGENT = 2;
	
	public int putAgent( T2DAgent agent, int x, int y )
	{
		int rval = AOK;
		Integer sp;
		
		// if putting a null (to empty a location), then, just do it.
		// otherwise check if something is already there first.
		if( (agent != null) )
		{
			// check space,
			//  only values of -1 are considered impenetrable in this model...
			if( ((sp = (Integer)currentSpace.getObjectAt( x, y )) != null) &&
				(sp.intValue() == -1 ) )
			{
				rval = ESPACE;
			}
			// then check the agent grid to see if the space is occupied
			else if( agentGrid.getObjectAt( x, y ) != null )
			{
				rval = EAGENT;
			}
			else
			{
				// AOK -- set the agent's actual location, with torus wrap
				agent.setXY( agentGrid.xnorm( x ), agentGrid.ynorm( y ) );
				// and mark the space...assume that 'sp' is valid
				// Rant: don't you wish that Number Objects were writable...
				sp = new Integer( sp.intValue() + 1 );
				// System.out.println( sp.intValue() );
				currentSpace.putObjectAt( x, y, sp );
			}
		}
		
		// put the agent (or null) in place on our grid...
		if( rval == AOK )
		{
			agentGrid.putObjectAt( x, y, agent );
		}
		
		return rval;
	}
	
	/** move specified agent to the specified location
	 * adjusts actual agent location for torus wrap-around
	 * check for already occupied destination
	 *  @return failures as in putAgent()
	 */
	public int moveAgent( T2DAgent agent, int x, int y )
	{
		int oldx = agent.getX();
		int oldy = agent.getY();
		
		// see if we succeed in putting the agent
		int rval = this.putAgent( agent, x, y );
		
		// remove agent from previous location
		if( rval == AOK )
			agentGrid.putObjectAt( oldx, oldy, null );
		
		return rval;
	}
	
	/**
	 * Place the specified agent in a random but un-occupied location.
	 *  Used to initialize agent positions
	 */
	public void placeAgent( T2DAgent agent )
	{
		int x, y;
		do
		{
			x = Uniform.staticNextIntFromTo( 0, size.width - 1 );
			y = Uniform.staticNextIntFromTo( 0, size.height - 1 );
		} while( this.putAgent( agent, x, y) != AOK );
	}
	
	/* Generate the display Objects for this kind of model.
	 * These are basically cannonical examples. There are other ways
	 * to go about generating displayable stuff should you want to
	 * complicate your life...
	 */
	
	/** Create a display to show the agentGrid on the screen.
	** This will display any Drawable contained within the Object2DTorus.
	**/
	public Object2DDisplay getAgentDisplay()
	{
		return new Object2DDisplay( agentGrid );
	}
	
	/** Create a color map of the currentSpace values to be displayed,
	 *   this is used by the Display to map values to colors without
	 *   having to call the space's Object.draw() method. Which is good
	 *   because the Objects are Integers that don't be Drawable...
	 */
	public Value2DDisplay getSpaceDisplay()
	{
		// make our map
		class myColorMap extends ColorMap
		{
			// extend and override in order to clamp map value at max
			myColorMap()
			{
				// map:
				// -1 to our icky-green obstacle indicator color
				// 0 to White...empty background
				// 64 to Black...full background
				this.mapColor( -1, new Color( 255 / 2, 255, 255 / 2 ) );
				this.mapColor( 0, Color.white );
				this.mapColor( 64, Color.black );

				// then map all the 1-63 ints to greys
				for( int i = 1; i < 64; ++i )
				{
					int v = 255 - 4 * i;
					this.mapColor( i, new Color( v, v, v ) );
				}
			}

			/**
			 * Gets the color mapped to the specified Integer
			 *  Rant: if the base-class getColor(Integer) had called gC(int)
			 *   this could have been easier....
			 */
			public Color getColor( Integer i )
			{
				return  this.getColor( i.intValue() );
			}

			/**
			 * Gets the color mapped to the specified int
			 *  and clamps at intval == 64
			 */
			public Color getColor( int i )
			{
				if( i > 64 )	i = 64;
				return  super.getColor( i );
			}

		};

		ColorMap map = new myColorMap();
	
		// currentSpace is an Object2DTorus containing Integer values at
		//  each x, y coordinate. Using the map created above this
		//  Value2DDisplay will show each of these Integers as a
		//  shade of grey, white if the Integer==0,
		//  or light-green if it is -1.
		return new Value2DDisplay( currentSpace, map );
	}
	
	// any other model specific code and variables here....

	
} // end'o'T2DSpace
